/*

*/

import renderer.scene.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;
import renderer.gui.*;

import java.awt.Color;
import java.awt.event.*;

/**


*/
@SuppressWarnings("serial")
public class InteractiveMaze_R18 extends InteractiveFrame
{
   private double fovy = 90.0;
   private boolean changePerspectiveVolumeWithNear = true;

   private boolean render_z_buffer = false;

   private boolean doBackfaceCulling = true;
   private boolean frontFacingIsCCW = true;

   private boolean letterbox = false;
   private double aspectRatio = 2.0;
   private double far    =  180.0;
   private double near   =  1.0;
   private boolean showCamera = false;

   private boolean showMatrix = false;

   private Scene scene;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
   */
   public InteractiveMaze_R18(String title, int fbWidth, int fbHeight)
   {
      super(title, fbWidth, fbHeight);

      // Create the Scene and Model objects that we shall render
      scene = new Scene();

      Model model = new Maze("maze2.txt");
      model.doBackfaceCulling = false;

      scene.addPosition(new Position(model));

      double fov    = 90.0;
      double aspect = 2.0;
      double near   = 1.0;
      double far    = 160.0;
      scene.camera.projPerspective(fov, aspect, near, far);

      // Middle of the maze, looking at the Mona Lisa.
      scene.camera.viewTranslate(80, 5, 80);
      scene.camera.viewRotate(180, 0, 1, 0);
   }


   // Implement part of the KeyListener interface.
   @Override public void keyPressed(KeyEvent e)
   {
      //System.out.println( e );

      int keyCode = e.getKeyCode();
      // Only handle the four arrow keys.
      if (KeyEvent.VK_UP == keyCode   || KeyEvent.VK_DOWN == keyCode
       || KeyEvent.VK_LEFT == keyCode || KeyEvent.VK_RIGHT == keyCode)
      {
         if (KeyEvent.VK_UP == keyCode && e.getModifiers() != 0 )
         {
         }
         else if (KeyEvent.VK_DOWN == keyCode && e.getModifiers() != 0 )
         {
         }
         else if (KeyEvent.VK_LEFT == keyCode && e.getModifiers() != 0 )
         {
         }
         else if (KeyEvent.VK_RIGHT == keyCode && e.getModifiers() != 0 )
         {
         }
         else if (KeyEvent.VK_UP == keyCode)
         {
            scene.camera.viewRotateX(1);
         }
         else if (KeyEvent.VK_DOWN == keyCode)
         {
            scene.camera.viewRotateX(-1);
         }
         else if (KeyEvent.VK_LEFT == keyCode)
         {
         }
         else if (KeyEvent.VK_RIGHT == keyCode)
         {
         }

         // Render again.
         setupViewport();
      }
   }


   // Implement part of the KeyListener interface.
   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('D' == c)
      {
         Pipeline.debug = ! Pipeline.debug;
       //Clip.debug = ! Clip.debug;
       //Rasterize.debug = ! Rasterize.debug;
      }
      else if (' ' == c)
      {
         render_z_buffer = ! render_z_buffer;
      }
      else if ('g' == c)
      {
         doBackfaceCulling = ! doBackfaceCulling;
         scene.positionList.get(0).model.doBackfaceCulling = doBackfaceCulling;
         System.out.print("Backface culling is ");
         System.out.println(doBackfaceCulling ? "on" : "off");
      }
      else if ('G' == c)
      {
         frontFacingIsCCW = ! frontFacingIsCCW;
         scene.positionList.get(0).model.frontFacingIsCCW = frontFacingIsCCW;
         System.out.print("Front face is ");
         System.out.println(frontFacingIsCCW ? "CCW" : "CW");
      }
      else if ('f' == c || 'F' == c)
      {
         // Move the camera's far plane.
         if ('f' == c)
         {
            far -= 0.01;
         }
         else
         {
            far += 0.01;
         }
      }
      else if ('n' == c || 'N' == c)
      {
         // Move the camera's near plane.
         if ('n' == c)
         {
            near -= 0.01;
         }
         else
         {
            near += 0.01;
         }
      }
      else if ('r' == c || 'R' == c)
      {
         // Change the aspect ratio of the camera's view rectangle.
         if ('r' == c)
         {
            aspectRatio -= 0.1;
         }
         else
         {
            aspectRatio += 0.1;
         }
      }
      else if ('o' == c || 'O' == c)
      {
         // Change the camera's field-of-view
         if ('o' == c)
         {
            fovy -= 1;
         }
         else
         {
            fovy += 1;
         }
      }
      else if ('l' == c)
      {
         letterbox = ! letterbox;
      }
      else if ('p' == c)
      {
         scene.camera.perspective = ! scene.camera.perspective;
      }
      else if ('M' == c)
      {
         showCamera = ! showCamera;
      }
      else if ('w' == c)
      {
         scene.camera.viewTranslate(0, 0, -1);
      }
      else if ('s' == c)
      {
         scene.camera.viewTranslate(0, 0, 1);
      }
      else if ('a' == c)
      {
         scene.camera.viewRotateY(1);
      }
      else if ('d' == c)
      {
         scene.camera.viewRotateY(-1);
      }
      else if ('j' == c)
      {
         scene.camera.viewTranslate(0, 1, 0);
      }
      else if ('k' == c)
      {
         scene.camera.viewTranslate(0, -1, 0);
      }

      // Set up the camera's view volume.
      if (scene.camera.perspective)
      {
         scene.camera.projPerspective(fovy, aspectRatio, near, far);
      }
      else
      {
         scene.camera.projOrtho(fovy, aspectRatio, near, far);
      }

      if (showCamera && ('M'==c
           ||'w'==c||'a'==c||'s'==c||'d'==c
           ||'j'==c||'k'==c
           ||'n'==c||'N'==c||'f'==c||'F'==c
           ||'o'==c||'O'==c||'r'==c||'R'==c||'p'==c))
      {
         System.out.print( scene.camera );
      }

      if ('m'==c)
      {
         System.out.print( scene.positionList.get(0).modelMatrix );
      }

      // Render again.
      setupViewport();
   }


   // Implement part of the ComponentListener interface.
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );

      // Get the new size of the FrameBufferPanel.
      int w = this.fbp.getWidth();
      int h = this.fbp.getHeight();

      // Create a new FrameBuffer that fits the new window size.
      FrameBuffer fb = new FrameBuffer(w, h);
      this.fbp.setFrameBuffer(fb);

      // Render again.
      setupViewport();
   }


   // Get in one place the code to set up the viewport.
   private void setupViewport()
   {
      // Render again.
      // Get the size of the FrameBuffer.
      FrameBuffer fb = this.fbp.getFrameBuffer();
      int w = fb.getWidthFB();
      int h = fb.getHeightFB();
      // Create a viewport with the correct aspect ratio.
      if ( letterbox )
      {
         if ( aspectRatio <= w/(double)h )
         {
            int width = (int)(h*aspectRatio);
            int xOffset = (w - width)/2;
            fb.setViewport(xOffset, 0, width, h);
         }
         else
         {
            int height = (int)(w/aspectRatio);
            int yOffset = (h - height)/2;
            fb.setViewport(0, yOffset, w, height);
         }
         fb.clearFB(Color.darkGray);
         fb.clearVP(new Color(135,206,235)); // skyblue
      }
      else // the viewport is the whole framebuffer
      {
         fb.clearFB(new Color(135,206,235)); // skyblue
         fb.setViewport();
      }
      Pipeline.render(scene, fb);
      if ( render_z_buffer )
      {
         FrameBuffer zBuf = this.fbp.getFrameBuffer().convertZB2FB();
         this.fbp.getFrameBuffer().setViewport(0, 0, zBuf);
      }
      fbp.update();
      repaint();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // Define initial dimensions for a FrameBuffer.
      int width  = 1024;
      int height = 512;
      // Create an InteractiveFrame containing a FrameBuffer
      // with the given dimensions. NOTE: We need to call the
      // InteractiveMaze_R18 constructor in the Java GUI Event
      // Dispatch Thread, otherwise we get a race condition
      // between the constructor (running in the main() thread)
      // and the very first ComponentEvent (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         new Runnable() // an anonymous inner class constructor
         {
            public void run() // implement the Runnable interface
            {
               // call the constructor that builds the gui
               new InteractiveMaze_R18("Renderer 18", width, height);
            }
         }
      );
   }//main()


   private static void print_help_message()
   {
      System.out.println("Use the 'D' key to toggle debugging information on and off.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the n/N keys to move the camera's near plane.");
      System.out.println("Use the f/F keys to move the camera's far plane.");
      System.out.println("Use the o/O keys to change the size of the camera's field-of-view.");
      System.out.println("Use the r/R keys to change the aspect ratio of the camera's view rectangle.");
      System.out.println("Use the 'l' key to toggle letterboxing on and off.");
      System.out.println("Use the arrow keys to move the camera location left/right and up/down.");
      System.out.println("Use CTRL arrow keys to rotate the camera left/right and up/down.");
      System.out.println("Use the 'm' key to toggle showing the Model transformation matrix.");
      System.out.println("Use the 'M' key to toggle showing the Camera's three matrices.");
      System.out.println("Use the 'g' key to toggle back face culling.");
      System.out.println("Use the 'G' key to toggle CW vs CCW.");
      System.out.println("Use the ' ' key to toggle between the image buffer and z-buffer.");
      System.out.println("Use the 'w' key to move forwards.");
      System.out.println("Use the 's' key to move backwards.");
      System.out.println("Use the 'a' key to turn left.");
      System.out.println("Use the 'd' key to turn right.");
      System.out.println("Use the 'k' key to move up.");
      System.out.println("Use the 'l' key to move down.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }
}
